"""Fuel Price helpers."""

import math

from geopy import (
    Nominatim,
    location,
    adapters
)

from ._version import __version__ as VERSION

async def geocode_reverse_lookup(coordinates: tuple) -> location.Location:
    """Reverse GeoCode lookup."""
    async with Nominatim(
        user_agent=f"pyfuelprices-{VERSION}",
        adapter_factory=adapters.AioHTTPAdapter) as geolocator:
        return await geolocator.reverse(coordinates, exactly_one=True, timeout=15, addressdetails=True)

class BoundingBox(object):
    """Represent a bounding box."""
    lat_min = None
    lon_min = None
    lat_max = None
    lon_max = None

def get_bounding_box(latitude_in_degrees, longitude_in_degrees, half_side_in_miles):
    """Return a bounding box from lat/long/radius"""
    assert half_side_in_miles > 0
    assert latitude_in_degrees >= -90.0 and latitude_in_degrees  <= 90.0
    assert longitude_in_degrees >= -180.0 and longitude_in_degrees <= 180.0

    half_side_in_km = half_side_in_miles * 1.609344
    lat = math.radians(latitude_in_degrees)
    lon = math.radians(longitude_in_degrees)

    radius  = 6371
    # Radius of the parallel at given latitude
    parallel_radius = radius*math.cos(lat)

    lat_min = lat - half_side_in_km/radius
    lat_max = lat + half_side_in_km/radius
    lon_min = lon - half_side_in_km/parallel_radius
    lon_max = lon + half_side_in_km/parallel_radius
    rad2deg = math.degrees

    box = BoundingBox()
    box.lat_min = rad2deg(lat_min)
    box.lon_min = rad2deg(lon_min)
    box.lat_max = rad2deg(lat_max)
    box.lon_max = rad2deg(lon_max)

    return box



"""Data sources and translators for pyfuelprices."""

import logging
import json

from datetime import timedelta, datetime
from typing import final
from geopy import (
    distance,
    Nominatim,
    location,
    adapters,
    exc as geopyexc
)
import aiohttp

from pyfuelprices.const import (
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_AREA_LAT,
    PROP_AREA_LONG,
    PROP_AREA_RADIUS
)
from pyfuelprices.fuel_locations import FuelLocation, Fuel
from pyfuelprices.helpers import geocode_reverse_lookup

_LOGGER = logging.getLogger(__name__)

class Source:
    """Base source, all instances inherit this."""

    _url: str = ""
    _method: str = "GET"
    _request_body: dict | None = None
    _headers: dict = {}
    _raw_data = None
    _timeout: int = 30
    _configured_areas: list[dict] = []
    _client_session: aiohttp.ClientSession = None
    update_interval: timedelta = timedelta(days=1)
    next_update: datetime = datetime.now()
    provider_name: str = ""
    location_cache: dict[str, FuelLocation] = None

    def __init__(self,
                 update_interval: timedelta = timedelta(days=1),
                 client_session: aiohttp.ClientSession = None) -> None:
        """Start a new instance of a source."""
        self.update_interval = update_interval
        self._client_session: aiohttp.ClientSession = client_session
        if client_session is None:
            self._client_session = aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(total=self._timeout)
            )
        else:
            self._client_session = client_session
        if self.next_update is None:
            self.next_update = datetime.now()

    @final
    def _check_if_coord_in_area(self, coordinates) -> bool:
        """Check to see if given coordinates are in a configured area."""
        for area in self._configured_areas:
            if distance.distance(
                coordinates,
                (area[PROP_AREA_LAT], area[PROP_AREA_LONG])
            ).miles <= area[PROP_AREA_RADIUS]:
                return True
        return False

    async def get_site(self, site_id) -> FuelLocation:
        """Return an individual fuel location from the remote API."""
        await self.location_cache[site_id].dynamic_build_fuels()
        return self.location_cache[site_id]

    async def search_sites(self, coordinates, radius: float) -> list[dict]:
        """Return all available sites within a given radius."""
        locations = []
        for site in self.location_cache.values():
            dist = distance.distance(coordinates,
                                (
                                    site.lat,
                                    site.long
                                )).miles
            if dist < radius:
                await site.dynamic_build_fuels()
                locations.append(
                    {
                        **site.__dict__(),
                        "distance": dist
                    }
                )
        return locations

    async def update(self, areas=None, force=False) -> list[FuelLocation]:
        """Update hooks for the data source."""
        _LOGGER.debug("Starting update hook for %s to url %s", self.provider_name, self._url)
        self._configured_areas=[] if areas is None else areas
        self._clear_cache()
        if self.next_update <= datetime.now() or force:
            response = await self._client_session.request(
                method=self._method,
                url=self._url,
                json=self._request_body,
                headers=self._headers
            )
            _LOGGER.debug("Update request completed for %s with status %s",
                        self.provider_name, response.status)
            if response.status == 200:
                self.next_update = datetime.now() + self.update_interval
                if "application/json" not in response.content_type:
                    return await self.parse_response(
                        response=json.loads(await response.text())
                    )
                return await self.parse_response(
                    response=await response.json()
                )
            raise UpdateFailedError(
                status=response.status,
                response=await response.text(),
                headers=response.headers
            )

    async def parse_response(self, response) -> list[FuelLocation]:
        """Parses the response from the update hook."""
        raise NotImplementedError("This function is not available for this module.")

    @final
    def _clear_cache(self):
        """Remove data for fuel stations that have not been accessed."""
        for x in range(len(self.location_cache)):
            last_access = list(self.location_cache.values())[x].last_access
            if last_access is not None:
                if last_access <= datetime.now()-timedelta(days=2):
                    if not list(self.location_cache.values())[x].props.get(
                        PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
                        False
                    ):
                        list(self.location_cache.values())[x].available_fuels = []

    def parse_fuels(self, fuels) -> list[Fuel]:
        """Parses the fuels response from the update hook.
        This is used as part of parse_response."""
        raise NotImplementedError("This function is not available for this module.")


class UpdateFailedError(Exception):
    """Update failure exception."""

    def __init__(self, status: int, response: str, headers: dict) -> None:
        self.status = status
        self.response = response
        self.headers = headers

class ServiceBlocked(UpdateFailedError):
    """Service Blocked exception."""



"""Netherlands data const."""

API_BASE = "https://tankservice.app-it-up.com/Tankservice/v2"

DIRECTLEASE_API_PLACES = f"{API_BASE}/places?fmt=web&country=NL&country=BE&lang=en"
DIRECTLEASE_API_STATION = API_BASE + "/places/{station_id}?_v48&lang=en"



"""Sources mapping file."""

from .austria import SOURCE_MAP as AT_SOURCE_MAP
from .australia import SOURCE_MAP as AU_SOURCE_MAP
from .germany import SOURCE_MAP as DE_SOURCE_MAP
from .greece import SOURCE_MAP as GR_SOURCE_MAP
from .romania import SOURCE_MAP as RO_SOURCE_MAP
from .uk.map import SOURCE_MAP as UK_SOURCE_MAP
from .usa import SOURCE_MAP as US_SOURCE_MAP
from .netherlands import SOURCE_MAP as NL_SOURCE_MAP

SOURCE_MAP = {
    **UK_SOURCE_MAP,
    **NL_SOURCE_MAP,
    **US_SOURCE_MAP,
    **DE_SOURCE_MAP,
    **AT_SOURCE_MAP,
    **AU_SOURCE_MAP,
    **RO_SOURCE_MAP,
    **GR_SOURCE_MAP
}

COUNTRY_MAP = {
    "BE": ["directlease"],
    "DE": [k for k in DE_SOURCE_MAP],
    "NL": [k for k in NL_SOURCE_MAP],
    "GB": [k for k in UK_SOURCE_MAP],
    "US": [k for k in US_SOURCE_MAP],
    "AU": [k for k in AU_SOURCE_MAP],
    "AT": [k for k in AT_SOURCE_MAP],
    "RO": [k for k in RO_SOURCE_MAP],
    "GR": [k for k in GR_SOURCE_MAP]
}



"""FuelGR Service for Greece."""

import logging
import asyncio

from datetime import timedelta, datetime

import xmltodict
from geopy import distance

from pyfuelprices.const import (
    PROP_AREA_LAT,
    PROP_AREA_LONG,
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE_ID
)
from pyfuelprices.sources import Source, geocode_reverse_lookup, geopyexc
from pyfuelprices.fuel_locations import Fuel, FuelLocation

from .const import (
    FUELGR_GET_DATA,
    FUELGR_GET_STATION_PRICES,
    FUELGR_FUEL_TYPE_MAPPING
)

_LOGGER = logging.getLogger(__name__)

class FuelGrSource(Source):
    """The root FuelGR data source."""

    update_interval = timedelta(days=1) # update once per day to prevent API spam.
    provider_name = "fuelgr"
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None
    _parser_coords = None

    async def _send_request(self, url) -> str:
        """Send a request to the API and return the raw response."""
        _LOGGER.debug("Sending HTTP request to FuelGR with URL %s", url)
        async with self._client_session.get(url=url,
                                            headers=self._headers) as response:
            if response.ok:
                return await response.text()
            _LOGGER.error("Error sending request to %s: %s",
                            url,
                            response)

    async def get_site(self, site_id) -> FuelLocation:
        """Return a single site."""
        return self.location_cache[site_id]

    async def search_sites(self, coordinates, radius: float) -> list[dict]:
        """Return all available sites (radius not used)"""
        # first query the API to populate the cache / update data
        await self.update_area(
            {
                PROP_AREA_LAT: coordinates[0],
                PROP_AREA_LONG: coordinates[1]
            }
        )
        locations = []
        for site in self.location_cache.values():
            dist = distance.distance(coordinates,
                                 (
                                    site.lat,
                                    site.long
                                )).miles
            if dist < radius:
                locations.append({
                    **site.__dict__(),
                    "distance": dist
                })
        return locations

    async def update_area(self, area: dict):
        """Used by asyncio to update areas."""
        try:
            geocode = await geocode_reverse_lookup(
                (area[PROP_AREA_LAT], area[PROP_AREA_LONG])
            )
        except geopyexc.GeocoderTimedOut:
            _LOGGER.warning("Timeout occured while geocoding area %s.",
                            area)
            return None
        if geocode.raw["address"]["country_code"] != "gr":
            _LOGGER.debug("Skipping area %s as not in GR.",
                        area)
            return None
        self._parser_coords = (area[PROP_AREA_LAT], area[PROP_AREA_LONG])
        _LOGGER.debug("Searching FuelGR for FuelLocations at area %s",
                    area)
        response_raw = await self._send_request(
            url=FUELGR_GET_DATA.format(
                LAT=area[PROP_AREA_LAT],
                LONG=area[PROP_AREA_LONG],
                FUEL="1",
                BRAND="0",
                QUALITY="0"
            )
        )
        # convert to dict
        response_raw = xmltodict.parse(response_raw)
        # now extract fuel stations
        response_raw = response_raw["gss"]["gs"]
        _LOGGER.debug("Found %s fuel stations",
                    len(response_raw))
        
        await self.parse_response(
            response=response_raw
        )

    async def update(self, areas=None, force=None) -> list[FuelLocation]:
        """Update the cached data."""
        if self.next_update <= datetime.now() or force:
            self._configured_areas=[] if areas is None else areas
            try:
                self.next_update += self.update_interval
                coros = [self.update_area(a) for a in self._configured_areas]
                asyncio.gather(*coros)
            except Exception as exc:
                _LOGGER.error(exc)

        return list(self.location_cache.values())

    async def get_fuel_station_fuels(self, station_id) -> list[Fuel]:
        """Return a list of fuels from a fuel station."""
        response_raw = await self._send_request(
            url=FUELGR_GET_STATION_PRICES.format(
                ID=station_id
            )
        )
        # convert to dict
        response_raw = xmltodict.parse(response_raw)
        response_raw = response_raw["gsf"]["fuel"]
        _LOGGER.debug("Found %s fuels for stn %s",
                      len(response_raw),
                      station_id)
        return self.parse_fuels(response_raw)

    async def parse_raw_fuel_station(self, station: dict) -> FuelLocation:
        """Parse a single raw fuel station into FuelLocation."""
        _LOGGER.debug("Parsing object %s", station["@guid"])
        site_id = f"{self.provider_name}_{station['@id']}"
        loc = FuelLocation.create(
            site_id=site_id,
            name=station["br"]["#text"],
            address=station["ad"],
            lat=station["lt"],
            long=station["lg"],
            brand=station["br"]["#text"],
            postal_code="",
            available_fuels=await self.get_fuel_station_fuels(
                station_id=station["@id"]
            ),
            currency="EUR",
            props={
                "data": station,
                PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True,
                PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                PROP_FUEL_LOCATION_SOURCE_ID: station["@id"]
            },
            next_update=self.next_update
        )
        if site_id not in self.location_cache:
            self.location_cache[site_id] = loc
        else:
            await self.location_cache[site_id].update(loc)
        return self.location_cache[site_id]

    async def parse_response(self, response) -> list[FuelLocation]:
        """Convert fuel stations into fuel objects."""
        coros = [self.parse_raw_fuel_station(s) for s in response]
        await asyncio.gather(*coros)

    def parse_fuels(self, fuels: dict) -> list[Fuel]:
        """Parse fuels using mapping"""
        fuel_resp = []
        for f in fuels:
            fuel_resp.append(
                Fuel(
                    fuel_type=FUELGR_FUEL_TYPE_MAPPING.get(f["@type"], f["fn"]),
                    cost=float(f["pr"]),
                    props=f
                )
            )

        return fuel_resp



"""Hold source mappings for modules."""

from pyfuelprices.sources import Source
from .applegreenstores import ApplegreenUKSource
from .ascona import AsconaGroupUKSource
from .asda import AsdaUKSource
from .bpuk import BpUKSource
from .essouk import EssoUKSource
from .jet import JetUKSource
from .morrisons import MorrisonsUKSource
from .moto import MotowayUKSource
from .motorfuelgroup import MotorFuelGroupUKSource
from .rontec import RontecUKSource
from .sainsburys import SainsburysUKSource
from .sgn import SgnRetailUKSource
from .shell import ShellUKSource
from .tesco import TescoUKSource

SOURCE_MAP: dict[str, Source] = {
    "applegreen": ApplegreenUKSource,
    "asda": AsdaUKSource,
    "ascona": AsconaGroupUKSource,
    "bpuk": BpUKSource,
    "essouk": EssoUKSource,
    "jet": JetUKSource,
    "morrisons": MorrisonsUKSource,
    "motoway": MotowayUKSource,
    "motorfuelgroup": MotorFuelGroupUKSource,
    "rontec": RontecUKSource,
    "sainsburys": SainsburysUKSource,
    "sgnretail": SgnRetailUKSource,
    "shelluk": ShellUKSource,
    "tesco": TescoUKSource
}



"""pyfuelprices locations."""

from typing import final
from datetime import datetime

from .const import PROP_FUEL_LOCATION_DYNAMIC_BUILD
from .fuel import Fuel

class FuelLocation:
    """Represents an invidual location."""

    _id = ""
    _name = ""
    _address = ""
    lat = 0.0
    long = 0.0
    _brand = ""
    available_fuels: list[Fuel] = []
    _currency = ""
    _postal_code: str | None = None
    last_updated: datetime | None = None
    next_update: datetime | None = None
    last_access: datetime | None = None
    props: dict = {}

    @final
    @property
    def id(self) -> str:
        """Return site_id."""
        self.last_access = datetime.now()
        return self._id

    @final
    @id.setter
    def id(self, new_val):
        """Set site_id"""
        self._id = new_val

    @final
    @property
    def name(self) -> str:
        """Return site name."""
        self.last_access = datetime.now()
        return self._name

    @final
    @name.setter
    def name(self, new_val):
        """Set site name."""
        self._name = new_val


    @final
    @property
    def address(self) -> str:
        """Return site address."""
        self.last_access = datetime.now()
        return self._address

    @final
    @address.setter
    def address(self, new_val):
        """Set site address."""
        self._address = new_val

    @final
    @property
    def brand(self) -> str:
        """Return fuel brand."""
        self.last_access = datetime.now()
        return self._brand

    @final
    @brand.setter
    def brand(self, new_val):
        """Set fuel brand."""
        self._brand = new_val

    @final
    @property
    def currency(self) -> str:
        """Return site name."""
        self.last_access = datetime.now()
        return self._currency

    @final
    @currency.setter
    def currency(self, new_val):
        """Set site currency."""
        self._currency = new_val

    @final
    @property
    def postal_code(self) -> str:
        """Return site postal code."""
        self.last_access = datetime.now()
        return self._postal_code

    @final
    @postal_code.setter
    def postal_code(self, new_val):
        """Set site postal code."""
        self._postal_code = new_val

    @final
    def __dict__(self) -> dict:
        """Convert the object to a dict."""
        fuels = {}
        for fuel in self.available_fuels:
            fuels[fuel.fuel_type] = fuel.cost
        return {
            "id": self._id,
            "name": self._name,
            "address": self._address,
            "postal_code": self._postal_code,
            "latitude": self.lat,
            "longitude": self.long,
            "brand": self._brand,
            "available_fuels": fuels,
            "currency": self._currency,
            "last_updated": self.last_updated.isoformat(),
            "next_update": (
                self.next_update.isoformat() if self.next_update is not None
                else "unavailable"),
            "props": self.props
        }

    @final
    async def update(self, updated: 'FuelLocation'):
        """Update the object."""
        self._name = updated.name
        self._address = updated.address
        self.lat = updated.lat
        self.long = updated.long
        self._brand = updated.brand
        self.last_updated = datetime.now()
        self._postal_code = updated.postal_code
        self.next_update = updated.next_update

        if (
            updated.props.get(PROP_FUEL_LOCATION_DYNAMIC_BUILD, False)
            and len(self.available_fuels)>0):
            await updated.dynamic_build_fuels()
        for fuel in updated.available_fuels:
            try:
                self.get_fuel(fuel.fuel_type).update(fuel.fuel_type, fuel.cost, fuel.props)
            except ValueError:
                self.available_fuels.append(fuel)

    @final
    def get_fuel(self, f_type: str) -> Fuel:
        """Return a fuel instance."""
        for fuel in self.available_fuels:
            if fuel.fuel_type == f_type:
                return fuel
        raise ValueError(f"No existing fuel data found for {f_type}")

    @final
    async def async_get_fuel(self, f_type: str) -> Fuel:
        """Return a fuel instance."""
        if self.props.get(PROP_FUEL_LOCATION_DYNAMIC_BUILD, False):
            await self.dynamic_build_fuels()
        for fuel in self.available_fuels:
            if fuel.fuel_type == f_type:
                return fuel
        raise ValueError(f"No existing fuel data found for {f_type}")

    async def dynamic_build_fuels(self):
        """Dynamic build of fuels for when accessing this data would normally be costly."""
        return True

    @classmethod
    def create(cls,
               site_id: str,
               name: str,
               address: str,
               lat: float,
               long: float,
               brand: str,
               available_fuels,
               last_updated: datetime = datetime.now(),
               next_update: datetime = None,
               postal_code: str | None = None,
               currency: str = "",
               props: dict = None
        ) -> 'FuelLocation':
        """Create an instance of fuel location."""
        location = cls()
        location._id = site_id
        location._address = address
        location._name = name
        location.lat = lat
        location.long = long
        location._brand = brand
        location.available_fuels = available_fuels
        location._postal_code = postal_code
        location._currency = currency
        location.last_updated = last_updated
        location.props = props
        location.next_update = next_update
        return location



"""Motor Fuel Group UK dataprovider."""

from pyfuelprices.sources.uk import CMAParser, FuelLocation

class MotowayUKSource(CMAParser):
    """Moto-Way UK uses the CMA parser."""

    _url = "https://moto-way.com/fuel-price/fuel_prices.json"
    provider_name = "motoway"
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None



"""UK Fuel Sources and parsers."""

from pyfuelprices.const import (
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE_ID
)
from pyfuelprices.sources import Source
from pyfuelprices.fuel import Fuel
from pyfuelprices.fuel_locations import FuelLocation

class CMAParser(Source):
    """This parser is specific for the scheme by the CMA."""

    location_cache: dict[str, FuelLocation] = {}

    async def parse_response(self, response) -> list[FuelLocation]:
        """Converts CMA data into fuel price mapping."""
        for location_raw in response["stations"]:
            site_id = f"{self.provider_name}_{location_raw['site_id']}"
            location = FuelLocation.create(
                    site_id=site_id,
                    name=f"{location_raw['brand']} {location_raw['postcode']}",
                    address=location_raw["address"],
                    brand=location_raw["brand"],
                    lat=location_raw["location"]["latitude"],
                    long=location_raw["location"]["longitude"],
                    available_fuels=self.parse_fuels(location_raw["prices"]),
                    postal_code=location_raw["postcode"],
                    currency="GBP",
                    props={
                        PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                        PROP_FUEL_LOCATION_SOURCE_ID: location_raw["site_id"],
                        PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True
                    },
                    next_update=self.next_update
                )
            if site_id not in self.location_cache:
                self.location_cache[site_id] = location
            else:
                await self.location_cache[site_id].update(location)
        return list(self.location_cache.values())

    def parse_fuels(self, fuels) -> list[Fuel]:
        """Parses fuel data."""
        f_list = []
        for f_type in fuels:
            if fuels[f_type] is not None:
                if fuels[f_type] > 10000:
                    fuels[f_type] = round(fuels[f_type]/100, 1)
                if fuels[f_type] > 1000:
                    fuels[f_type] = round(fuels[f_type]/10, 1)
                if fuels[f_type] < 100:
                    fuels[f_type] = round(fuels[f_type]*100, 1)
            f_list.append(Fuel(
                fuel_type=f_type,
                cost=fuels[f_type],
                props={}
            ))
        return f_list



"""Applegreen Stores dataprovider."""

from pyfuelprices.sources.uk import CMAParser, FuelLocation

class ApplegreenUKSource(CMAParser):
    """Applegreen UK uses the CMA parser."""

    _url = "https://applegreenstores.com/fuel-prices/data.json"
    provider_name = "Applegreen"
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None



"""Ascona Group UK dataprovider."""

from pyfuelprices.sources.uk import CMAParser, FuelLocation

class AsconaGroupUKSource(CMAParser):
    """Asconagrounp UK uses the CMA parser."""

    _url = "https://fuelprices.asconagroup.co.uk/newfuel.json"
    provider_name = "ascona"
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None



"""PecoOnline Service for Romania."""

import logging
import uuid
import json
from datetime import timedelta

from pyfuelprices.const import (
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE_ID
)
from pyfuelprices.sources import Source
from pyfuelprices.fuel_locations import Fuel, FuelLocation

from .const import (
    PECO_STANDARD_HEADERS,
    PECO_ONLINE_API,
    PECO_QUERY_CLAUSE,
    PECO_FUEL_MAPPING
)

_LOGGER = logging.getLogger(__name__)

class PecoOnlineSource(Source):
    """The root Peco Online data source."""

    _url = PECO_ONLINE_API
    update_interval = timedelta(days=1) # update once per day to prevent API spam.
    provider_name = "pecoonline"
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None
    _headers = PECO_STANDARD_HEADERS

    async def _send_request(self, url, body) -> str:
        """Send a request to the API and return the raw response."""
        _LOGGER.debug("Sending HTTP request to PecoOnline with URL %s", url)
        if self._headers["X-Parse-Installation-Id"] is None:
            self._headers["X-Parse-Installation-Id"] = str(uuid.uuid4())
        async with self._client_session.post(url=url,
                                             json=body,
                                            headers=self._headers) as response:
            if response.ok:
                return await response.text()
            _LOGGER.error("Error sending request to %s: %s",
                            url,
                            response)

    async def get_site(self, site_id) -> FuelLocation:
        """Return a single site."""
        return self.location_cache[site_id]

    async def update(self, areas=None, force=None) -> list[FuelLocation]:
        """Update the cached data."""
        try:
            response_raw = await self._send_request(
                url=self._url,
                body={
                    "limit": 100000,
                    "where": PECO_QUERY_CLAUSE,
                    "order": "Benzina_Regular",
                    "_method": "GET"
                }
            )
            await self.parse_response(
                response=json.loads(response_raw)
            )
        except Exception as exc:
            _LOGGER.error(exc)

        return list(self.location_cache.values())

    async def parse_raw_fuel_station(self, station: dict) -> FuelLocation:
        """Parse a single raw fuel station into FuelLocation."""
        site_id = f"{self.provider_name}_{station['objectId']}"
        loc = FuelLocation.create(
            site_id=site_id,
            name=station["Statie"],
            address=station["Adresa"],
            lat=station["lat"],
            long=station["lng"],
            brand=station["Retea"],
            postal_code="See address",
            available_fuels=self.parse_fuels(station),
            currency="RON",
            props={
                PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True,
                PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                PROP_FUEL_LOCATION_SOURCE_ID: station['objectId']
            }
        )
        if site_id not in self.location_cache:
            self.location_cache[site_id] = loc
        else:
            await self.location_cache[site_id].update(loc)
        return self.location_cache[site_id]

    async def parse_response(self, response) -> list[FuelLocation]:
        """Convert fuel stations into fuel objects."""
        for result in response["results"]:
            _LOGGER.debug("Parsing object %s", result["objectId"])
            await self.parse_raw_fuel_station(result)

    def parse_fuels(self, fuels: dict) -> list[Fuel]:
        """Parse fuels using mapping"""
        fuel_resp = []
        for k, v in PECO_FUEL_MAPPING.items():
            if k in fuels:
                fuel_resp.append(
                    Fuel(
                        fuel_type=v,
                        cost=fuels[k],
                        props=None
                    )
                )

        return fuel_resp



"""Morrisons UK dataprovider."""

from pyfuelprices.const import DESKTOP_USER_AGENT
from pyfuelprices.sources.uk import CMAParser, FuelLocation

class MorrisonsUKSource(CMAParser):
    """Morrisons UK uses the CMA parser."""

    _url = "https://www.morrisons.com/fuel-prices/fuel.json"
    provider_name = "morrisons"
    _headers = {
        "User-Agent": DESKTOP_USER_AGENT,
        "Accept": "application/json",
        "Accept-Language": "en-GB,en;q=0.5",
        "Accept-Encoding": "gzip, deflate, br",
        "Connection": "keep-alive",
        "Upgrade-Insecure-Requests": "1",
        "Sec-Fetch-Dest": "document",
        "Sec-Fetch-Mode": "navigate",
        "Sec-Fetch-Site": "cross-site",
        "Sec-Fetch-User": "?1"
    }
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None



"""Greece data const."""

FUELGR_API_DEV = "android.4.0-b2da2cf97330ca3b"
FUELGR_API_SIG = "google/coral/coral:14/UQ1A.240205.004/1709778835:userdebug/release-keys"
FUELGR_API_APKSIG = "UPJ2YQunu9eGXu8a/WOiVNAZlYA="

FUELGR_GET_DATA = (
    "https://deixto.gr/fuel/get_data_v4.php?"
    f"dev={FUELGR_API_DEV}"
    "&lat={LAT}"
    "&long={LONG}"
    "&f={FUEL}"
    "&b={BRAND}"
    "&d=30"
    "&p={QUALITY}"
    f"&dSig={FUELGR_API_SIG}"
    "&iLoc=unknown"
    f"&apkSig={FUELGR_API_APKSIG}"
)
FUELGR_GET_STATION_PRICES = (
    "https://deixto.gr/fuel/get_gasstation_prices.php?"
    f"dev={FUELGR_API_DEV}"
    "&gsid={ID}"
)

FUELGR_FUEL_TYPE_MAPPING = {
    "1": "95",
    "2": "98-100",
    "3": "Super",
    "4": "Diesel",
    "5": "Heating Diesel",
    "6": "LPG",
    "7": "Home Heating Diesel",
    "8": "CNG"
}

FUELGR_BRAND_MAPPING = {
    "0": "ALL",
    "1": "ΑΝΕΞΑΡΤΗΤΟ ΠΡΑΤΗΡΙΟ",
    "2": "JETOIL",
    "3": "SHELL",
    "5": "ETEKA",
    "6": "AVIN",
    "8": "EKO",
    "9": "REVOIL",
    "10": "BP",
    "11": "ΕΛΙΝΟΙΛ",
    "13": "CYCLON",
    "17": "KAOIL"
}



"""SGN Retail UK dataprovider."""

from pyfuelprices.fuel import Fuel
from pyfuelprices.sources.uk import CMAParser, FuelLocation

class SgnRetailUKSource(CMAParser):
    """SGN Retail UK uses the CMA parser."""

    _url = "https://www.sgnretail.uk/files/data/SGN_daily_fuel_prices.json"
    provider_name = "sgnretail"
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None



"""Esso UK module."""

from pyfuelprices.sources.uk import CMAParser, FuelLocation

class EssoUKSource(CMAParser):
    """Esso UK uses the same parser as CMA."""

    _url = "https://fuelprices.esso.co.uk/latestdata.json"
    provider_name = "essouk"
    _timeout = 60
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None



"""GasBuddy USA data source."""
import logging
import json
import uuid

# use these-united-states for geocoding to states
import united_states
from geopy import distance

from pyfuelprices.fuel_locations import Fuel, FuelLocation

from pyfuelprices.const import (
    PROP_AREA_LAT,
    PROP_AREA_LONG,
    PROP_AREA_RADIUS,
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE_ID,
    ANDROID_USER_AGENT
)
from pyfuelprices.sources import Source

_LOGGER = logging.getLogger(__name__)

CONST_GASBUDDY_STATIONS_FMT = (
    "https://services.gasbuddy.com/mobile-orchestration/stations?authid={AUTHID}"
    "&country={COUNTRY}&distance_format={DISTANCEFMT}&limit={LIMIT}&region={STATE}"
    "&lat={LAT}&lng={LONG}"
)
CONST_GASBUDDY_GET_STATION_FMT = (
    "https://services.gasbuddy.com/mobile-orchestration/stations/{STATIONID}"
    "?authid={AUTHID}"
)

class GasBuddyUSASource(Source):
    """GasBuddy USA data source."""

    _headers = {
        "apikey": "56c57e8f1132465d817d6a753c59387e",
        "User-Agent": ANDROID_USER_AGENT
    }
    _usa = united_states.UnitedStates()
    _parser_radius = 0
    _parser_coords = ()
    provider_name = "gasbuddy"
    location_cache = {}

    async def _send_request(self, url) -> str:
        """Send a request to the API and return the raw response."""
        _LOGGER.debug("Sending HTTP request to GasBuddy with URL %s", url)
        async with self._client_session.get(url=url,
                                            headers=self._headers) as response:
            if response.ok:
                return await response.text()
            _LOGGER.error("Error sending request to %s: %s",
                            url,
                            response)

    async def get_site(self, site_id) -> FuelLocation:
        """Return the site from the location cache."""
        if site_id in self.location_cache:
            return self.location_cache[site_id]
        if "_" in str(site_id):
            site_id = site_id.split("_")
        response_raw = await self._send_request(
            url=CONST_GASBUDDY_GET_STATION_FMT.format(
                STATIONID=str(site_id),
                AUTHID=str(uuid.uuid4)
            )
        )
        return await self.parse_raw_fuel_station(
            station=json.loads(response_raw)["station"]
        )

    async def search_sites(self, coordinates, radius: float) -> list[dict]:
        """Return all available sites within a given radius."""
        # first query the API to populate cache / update data in case this data is unavailable.
        await self.update(
            areas=[{
                PROP_AREA_LAT: coordinates[0],
                PROP_AREA_LONG: coordinates[1],
                PROP_AREA_RADIUS: radius
            }]
        )
        locations = []
        for site in self.location_cache.values():
            dist = distance.distance(coordinates,
                                 (
                                    site.lat,
                                    site.long
                                )).miles
            if dist < radius:
                locations.append({
                    **site.__dict__(),
                    "distance": dist
                })
        return locations

    async def update(self, areas=None, force=None) -> list[FuelLocation]:
        """Custom update handler as this needs to query GasBuddy on areas."""
        self._configured_areas=[] if areas is None else areas
        for area in self._configured_areas:
            self._parser_coords = (area[PROP_AREA_LAT], area[PROP_AREA_LONG])
            self._parser_radius = area[PROP_AREA_RADIUS]
            coord_data = self._usa.from_coords(
                lat=area[PROP_AREA_LAT],
                lon=area[PROP_AREA_LONG]
            )
            if len(coord_data) > 0:
                _LOGGER.debug("Searching GasBuddy for FuelLocations at area %s",
                              area)
                coord_data = coord_data[0]
                response_raw = await self._send_request(
                    url=CONST_GASBUDDY_STATIONS_FMT.format(
                        AUTHID=str(uuid.uuid4()),
                        COUNTRY="US",
                        DISTANCEFMT="auto",
                        LIMIT=1000,
                        STATE=coord_data.abbr,
                        LAT=area[PROP_AREA_LAT],
                        LONG=area[PROP_AREA_LONG]
                    )
                )
                await self.parse_response(json.loads(response_raw))
        return list(self.location_cache.values())

    async def parse_raw_fuel_station(self, station) -> FuelLocation:
        """Converts a raw instance of a fuel station into a fuel location."""
        info = station["info"]
        site_id = f"{self.provider_name}_{station['id']}"
        _LOGGER.debug("Parsing GasBuddy location ID %s", site_id)
        loc = FuelLocation.create(
            site_id=site_id,
            name=f"{info['name']} {info['address']['postal_code']}",
            address="See properties",
            lat=info["latitude"],
            long=info["longitude"],
            brand=info["brand_name"],
            available_fuels=self.parse_fuels(station["fuel_products"]),
            postal_code=info['address']['postal_code'],
            currency="USD",
            props={
                "data": station,
                PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True,
                PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                PROP_FUEL_LOCATION_SOURCE_ID: station["id"]
            },
            next_update=self.next_update
        )
        if site_id not in self.location_cache:
            self.location_cache[site_id] = loc
        else:
            await self.location_cache[site_id].update(loc)
        return self.location_cache[site_id]

    async def parse_response(self, response) -> list[FuelLocation]:
        if response.get("stations", None) is not None:
            response = response["stations"]
        for station in response:
            info = station["info"]
            if distance.distance(
                self._parser_coords,
                (info["latitude"], info["longitude"])
            ).miles <= self._parser_radius:
                await self.parse_raw_fuel_station(station=station)

        return list(self.location_cache.values())

    def parse_fuels(self, fuels) -> list[Fuel]:
        """Parse fuels from fuel_products."""
        fuel_parsed = []
        for fuel in fuels:
            cost = fuel.get("cash",
                            fuel.get("credit", None))
            fuel_parsed.append(Fuel(
                fuel_type=fuel["fuel_product"],
                cost=0 if cost is None else cost["price"],
                props=fuel
            ))
        return fuel_parsed



"""Shell UK dataprovider."""

import logging
import json

from datetime import datetime
from pyfuelprices.const import DESKTOP_USER_AGENT
from pyfuelprices.sources import UpdateFailedError
from pyfuelprices.sources.uk import CMAParser, FuelLocation

_LOGGER = logging.getLogger(__name__)

class ShellUKSource(CMAParser):
    """Shell UK uses the CMA parser although requires custom request handlers."""

    _url = ("https://prodpricinghubstrgacct.blob.core.windows.net/ukcma/fuel-prices-data.json"
            "?sp=r&st=2023-11-21T05%3A20%3A29Z&se=2024-11-21T13%3A20%3A29Z&spr=https&"
            "sv=2022-11-02&sr=b&sig=%2F4eArYrrj1qKpD6Kn3a8on7Fm3jqTdBAKeH04gsuNho%3D")
    provider_name = "shelluk"
    _headers = {
        "User-Agent": DESKTOP_USER_AGENT
    }
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None

    async def update(self, areas=None, force=False) -> list[FuelLocation]:
        """Update hooks for the data source."""
        if datetime.now() > self.next_update or force:
            _LOGGER.debug("Starting update hook for %s to url %s", self.provider_name, self._url)
            async with self._client_session.request(
                url=self._url,
                method=self._method,
                json=self._request_body,
                headers=self._headers
            ) as response:
                _LOGGER.debug("Update request completed for %s with status %s",
                              self.provider_name, response.status)
                if response.status == 200:
                    self.next_update += self.update_interval
                    # convert octet stream to json
                    try:
                        js = json.loads(await response.text())
                        return await self.parse_response(
                            response=js
                        )
                    except Exception as exc:
                        raise UpdateFailedError(
                            status=response.status,
                            response="Invalid response provided.",
                            headers=response.headers
                        ) from exc
                raise UpdateFailedError(
                    status=response.status,
                    response=await response.text(),
                    headers=response.headers
                )



"""TankerKoenig data source."""

import logging
import json

from datetime import datetime

from pyfuelprices.const import (
    PROP_AREA_LAT,
    PROP_AREA_LONG,
    PROP_AREA_RADIUS,
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE_ID,
    DESKTOP_USER_AGENT
)
from pyfuelprices.fuel_locations import Fuel, FuelLocation
from pyfuelprices.sources import (
    Source,
    geocode_reverse_lookup,
    geopyexc
)

_LOGGER = logging.getLogger(__name__)

CONST_TANKERKOENIG_API_BASE = "https://tankerkoenig.de/ajax_v3_public/"
CONST_TANKERKOENIG_GET_STATIONS = (
    f"{CONST_TANKERKOENIG_API_BASE}get_stations_near_postcode.php"
    "?postcode={POSTCODE}&radius={RADIUS}"
)

class TankerKoenigSource(Source):
    """TankerKoenig Source."""
    _headers = {
        "User-Agent": DESKTOP_USER_AGENT
    }
    provider_name = "tankerkoenig"
    location_cache: dict[str, FuelLocation] = {}
    # location_tree = None

    async def _send_request(self, postcode, radius):
        """Send a request to the API for a given postcode and radius."""
        radius = radius * 1.609344 # convert miles to km
        url = CONST_TANKERKOENIG_GET_STATIONS.format(
            POSTCODE=postcode,
            RADIUS=radius
        )
        _LOGGER.debug("Sending request to TankerKoenig: %s",
                      url)
        async with self._client_session.get(
            url,
            headers=self._headers) as response:
            if response.ok:
                return await response.text()
            _LOGGER.error("Error sending request to %s: %s",
                            url,
                            response)

    async def update(self, areas=None, force=None) -> list[FuelLocation]:
        """Custom update handler as this needs to query TankerKoenig on areas."""
        if datetime.now() > self.next_update:
            self._configured_areas=[] if areas is None else areas
            for area in self._configured_areas:
                try:
                    geocode = await geocode_reverse_lookup(
                        (area[PROP_AREA_LAT], area[PROP_AREA_LONG])
                    )
                except geopyexc.GeocoderTimedOut:
                    _LOGGER.warning("Timeout occured while geocoding area %s.",
                                    area)
                    continue
                if geocode.raw["address"]["country_code"] != "de":
                    _LOGGER.debug("Skipping area %s as not in DE.",
                                area)
                    continue
                response_raw = json.loads(await self._send_request(
                    postcode=geocode.raw["address"]["postcode"],
                    radius=area[PROP_AREA_RADIUS]
                ))
                if response_raw["ok"]:
                    await self.parse_response(response_raw["data"])
                else:
                    _LOGGER.error("Error sending request to %s: %s",
                                self.provider_name,
                                response_raw)
            self.next_update += self.update_interval
        return list(self.location_cache.values())

    def _parse_raw(self, station: dict) -> FuelLocation:
        """Parse a single raw instance of a fuel site."""
        site_id = f"{self.provider_name}_{station['id']}"
        _LOGGER.debug("Parsing TankerKoenig location ID %s", site_id)
        loc = FuelLocation.create(
            site_id=site_id,
            name=station["name"],
            address=f"{station['house_number']} {station['street']} {station['post_code']}",
            lat=station["lat"],
            long=station["lng"],
            brand=station["brand"],
            available_fuels=self.parse_fuels(station["prices"]),
            postal_code=station['post_code'],
            currency="EUR",
            props={
                PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                PROP_FUEL_LOCATION_SOURCE_ID: station["id"],
                PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True,
                "opening_times": station.get("openingTimes", {}).get("openingTimes", [])
            }
        )
        loc.next_update = self.next_update if self.next_update > datetime.now() else (
            self.next_update + self.update_interval
        )
        return loc

    async def parse_response(self, response) -> list[FuelLocation]:
        for station_raw in response.get("stations", []):
            station = self._parse_raw(station_raw)
            if station.id not in self.location_cache:
                self.location_cache[station.id] = station
            else:
                await self.location_cache[station.id].update(station)
        return list(self.location_cache.values())

    def parse_fuels(self, fuels: dict[str, object]) -> list[Fuel]:
        """Parse fuel data from TankerKoenig."""
        fuels_parsed = []
        for k in fuels:
            if k.find("_") < 0:
                fuels_parsed.append(Fuel(
                    fuel_type=str(k).upper(),
                    cost=fuels[k],
                    props={}
                ))
        return fuels_parsed



"""DirectLease data source."""

import random
import re
import logging
import uuid
import hashlib
from datetime import datetime, timedelta
from time import time

import aiohttp

from pyfuelprices.const import (
    PROP_FUEL_LOCATION_DYNAMIC_BUILD,
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_SOURCE_ID,
    ANDROID_USER_AGENT
)
from pyfuelprices.sources import ServiceBlocked, Source, UpdateFailedError
from pyfuelprices.fuel_locations import Fuel, FuelLocation

from .const import DIRECTLEASE_API_PLACES, DIRECTLEASE_API_STATION

_LOGGER = logging.getLogger(__name__)


def _hash(hsh: str):
    encoded_string = hsh.encode("utf-8")
    hash_object = hashlib.sha1(encoded_string)
    return hash_object.hexdigest()

def _checksum_generator(url: str) -> str:
    """Generate a directlease API checksum."""
    # first generate a dummy device identifier
    date_string = datetime.now().strftime("%Y%m%d")
    device_uuid = str(uuid.uuid4())
    date_uuid_part = f"{date_string}_{device_uuid}"
    timestamp = int(time())
    parts = url.split("/")
    file_path = "/".join(parts[3:])
    file_path = f"/{file_path}"
    base_string = f"{date_uuid_part}/{timestamp}/{file_path}/X-Checksum"
    return f"{date_uuid_part}/{timestamp}/{_hash(base_string)}"

class DirectLeaseFuelLocation(FuelLocation):
    """DirectLease custom fuel location."""

    next_update: datetime = datetime.now()
    _client_session: aiohttp.ClientSession

    @classmethod
    def create(cls,
               site_id: str,
               name: str,
               address: str,
               lat: float,
               long: float,
               brand: str,
               available_fuels,
               last_updated: datetime = datetime.now(),
               postal_code: str | None = None,
               currency: str = "",
               props: dict = None,
               client_session: aiohttp.ClientSession = None
        ) -> 'FuelLocation':
        """Create an instance of fuel location."""
        location = cls()
        location._id = site_id
        location._address = address
        location._name = name
        location.lat = lat
        location.long = long
        location._brand = brand
        location.available_fuels = available_fuels
        location._postal_code = postal_code
        location._currency = currency
        location.last_updated = last_updated
        location.props = props
        location._client_session = client_session
        return location

    async def dynamic_build_fuels(self):
        """Dynamic requests to build fuels."""
        _LOGGER.debug("Update requested for station %s", self._id)
        if (len(self.available_fuels) != 0 and self.next_update > datetime.now()):
            _LOGGER.debug("Update skipped, next update required at %s, current fuel count %s",
                          self.next_update,
                          len(self.available_fuels))
            return None

        request_url = DIRECTLEASE_API_STATION.format(
            station_id=self.props[PROP_FUEL_LOCATION_SOURCE_ID]
        )

        _LOGGER.debug("Requesting fuel prices for station %s (%s)", self._id, request_url)

        try:
            async with self._client_session.request(
            method="GET",
            url=request_url,
            headers={
                "X-Checksum": _checksum_generator(request_url),
                "User-Agent": ANDROID_USER_AGENT
            }) as response:
                _LOGGER.debug("Got status code %s for dynamic parse of site %s",
                                response.status,
                                self.props[PROP_FUEL_LOCATION_SOURCE_ID])
                if response.status == 200:
                    response = await response.json()
                    # opportunity to build extra location data too
                    self._postal_code = response.get("postalCode", "Unknown")
                    self._address = response.get("address", "Unknown")
                    self._brand = response.get("brand", "Unknown")
                    self._name = response.get("name", "Unknown")
                    for fuel_raw in response["fuels"]:
                        _LOGGER.debug("Parsing fuel %s", fuel_raw)
                        f_type = re.search(r"\(([^)]+)\)", fuel_raw["name"])
                        if f_type is None:
                            f_type = fuel_raw["name"].upper()
                        else:
                            f_type = f_type.group(1).upper()
                        try:
                            fuel_cost = fuel_raw.get("price", None)
                            if fuel_cost is not None:
                                self.get_fuel(f_type).update(
                                    fuel_type=f_type,
                                    cost=fuel_raw["price"]/1000,
                                    props={}
                                )
                            else:
                                self.get_fuel(f_type).update(
                                    fuel_type=f_type,
                                    cost=0,
                                    props={
                                        "unavailable": True
                                    }
                                )
                        except ValueError:
                            if fuel_raw.get("price", None) is not None:
                                self.available_fuels.append(
                                    Fuel(
                                        fuel_type=f_type,
                                        cost=fuel_raw["price"]/1000,
                                        props={}
                                    )
                                )
                elif response.status == 403:
                    # 403 for directlease is an IP block
                    _LOGGER.error(
                        ("DirectLease appears to have blocked your IP. Your options are:"
                        "1) Contact tankservice-block@app-it-up.com for further support. "
                        "2) Change your WAN IP address or use a VPN. "
                        "3) Use a proxy server to connect. "))

            self.next_update = datetime.now() + timedelta(
            days=1,
            minutes=random.randint(1,240),
            seconds=random.randint(1,60)) # prevent overloading the API and IP lockouts

        except aiohttp.ServerDisconnectedError as err:
            _LOGGER.warning("Error sending request to URL %s: %s",
                            request_url,
                            err)

class DirectLeaseTankServiceParser(Source):
    """DirectLease parser for Belgium/Netherlands data."""

    _url = DIRECTLEASE_API_PLACES
    update_interval = timedelta(days=1) # update once per day to prevent API spam.
    provider_name = "directlease"
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None

    async def get_site(self, site_id) -> FuelLocation:
        await self.location_cache[site_id].dynamic_build_fuels()
        return self.location_cache[site_id]

    async def update(self, areas=None, force=None) -> list[FuelLocation]:
        try:
            await super().update(areas)
        except UpdateFailedError as err:
            if err.status == 403:
                # 403 for directlease is an IP block
                _LOGGER.error(("DirectLease appears to have blocked your IP. Your options are:"
                                "1) Contact tankservice-block@app-it-up.com for further support. "
                                "2) Change your WAN IP address or use a VPN. "
                                "3) Use a proxy server to connect. "))
                raise ServiceBlocked(err.status, err.response, err.headers) from err

        self.next_update = datetime.now() + timedelta(
            days=1,
            minutes=random.randint(1,240),
            seconds=random.randint(1,60)) # prevent overloading the API and IP lockouts
        return list(self.location_cache.values())

    async def parse_response(self, response) -> list[DirectLeaseFuelLocation]:
        """Fuel location parser for DirectLease."""
        for raw_loc in response:
            site_id = f"directlease_{raw_loc['id']}"
            if site_id not in self.location_cache:
                _LOGGER.debug("Parsing DirectLease location ID %s", site_id)
                loc = DirectLeaseFuelLocation.create(
                    site_id=site_id,
                    name=raw_loc.get("name", None),
                    address="Unknown",
                    lat=raw_loc["lat"],
                    long=raw_loc["lng"],
                    brand=raw_loc.get("brand", None),
                    available_fuels=[],
                    postal_code="Unknown",
                    currency="EUR",
                    props={
                        PROP_FUEL_LOCATION_DYNAMIC_BUILD: True,
                        PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                        PROP_FUEL_LOCATION_SOURCE_ID: raw_loc["id"]
                    },
                    last_updated=datetime.now(),
                    client_session=self._client_session
                )
                if (loc.name is None and loc.brand is not None):
                    loc.name = raw_loc["brand"] + " " + raw_loc["city"]
                if loc.name is None:
                    loc.name = f"Unknown site {loc.props[PROP_FUEL_LOCATION_SOURCE_ID]}"
                if loc.brand is None:
                    loc.brand = "Unknown"
                self.location_cache[site_id] = loc

        for site in self.location_cache.values():
            if (self._check_if_coord_in_area((
                    site.lat, site.long)) or (
                    len(site.available_fuels)>0
                )):
                await site.dynamic_build_fuels()

        return list(self.location_cache.values())

    def parse_fuels(self, fuels) -> list[Fuel]:
        """Fuel parser not required for this module."""



"""Representation of a single fuel."""

class Fuel:
    """Individual Fuel."""

    def __init__(self, fuel_type: str, cost: float, props: dict):
        """Initalize a fuel price."""
        self.fuel_type = fuel_type
        if cost is None:
            self.cost = 0
        else:
            self.cost = cost
        self.props = props

    def __dict__(self) -> dict:
        """Convert to a dict."""
        return {
            "type": self.fuel_type,
            "cost": self.cost,
            "properties": self.props
        }

    def update(self, fuel_type: str, cost: float, props: dict):
        """Update this instance of data."""
        self.fuel_type = fuel_type
        if cost is None:
            self.cost = 0
        else:
            self.cost = cost
        self.props = props



"""The core fuel prices module."""

import logging
import asyncio

from datetime import timedelta

import aiohttp

from pyfuelprices.sources import Source, geocode_reverse_lookup
from pyfuelprices.sources.mapping import SOURCE_MAP, COUNTRY_MAP
from .const import PROP_FUEL_LOCATION_SOURCE
from .fuel_locations import FuelLocation

_LOGGER = logging.getLogger(__name__)

class FuelPrices:
    """The base fuel prices entry class."""

    configured_sources: dict[str, Source] = {}
    configured_areas: list[dict] = []
    _accessed_sites: dict[str, str] = {}
    client_session: aiohttp.ClientSession = None

    async def update(self, force: bool=False):
        """Main data fetch / update handler."""
        async def update_src(s: Source, a: list[dict], f: bool):
            """Update source."""
            try:
                async with asyncio.Semaphore(4):
                    await s.update(areas=a, force=f)
            except TimeoutError as err:
                _LOGGER.warning("Timeout updating %s: %s", s.provider_name, err)
        coros = [
            update_src(s, self.configured_areas, force) for s in self.configured_sources.values()
        ]
        await asyncio.gather(*coros)

    async def get_fuel_location(self, site_id: str, source_id: str) -> FuelLocation:
        """Retrieve a single fuel location (supporting dynamic parse)."""
        if site_id not in self._accessed_sites:
            self._accessed_sites[site_id] = source_id
        return await self.configured_sources[source_id].get_site(site_id)

    async def find_fuel_locations_from_point(self,
                                       coordinates,
                                       radius: float,
                                       source_id: str = "") -> list[dict]:
        """Retrieve all fuel locations from a single point."""
        _LOGGER.debug("Searching for all fuel locations at point %s with a %s "
                      "mile radius for source %s.",
                      coordinates,
                      radius,
                      source_id if source_id != "" else "any")
        if source_id != "":
            return await self.configured_sources[source_id].search_sites(
                coordinates=coordinates,
                radius=radius
            )
        geocoded = (await geocode_reverse_lookup(coordinates)).raw['address']['country_code']
        if geocoded.upper() not in COUNTRY_MAP:
            raise ValueError("No data source exists for the given coordinates.", geocoded)
        locations = []
        for src in COUNTRY_MAP.get(geocoded.upper(), []):
            if src in self.configured_sources:
                locations.extend(await self.configured_sources[src].search_sites(
                    coordinates=coordinates,
                    radius=radius
                ))
        return locations

    async def find_fuel_from_point(self,
                             coordinates,
                             radius: float,
                             fuel_type: str,
                             source_id: str = "") -> list[dict]:
        """Retrieve the fuel cost from a single point."""
        async def dynamic_build(l: dict):
            """Function for asyncio to retrieve fuels quickly."""
            async with asyncio.Semaphore(5):
                await self.get_fuel_location(
                    l["id"],
                    str(l["props"]["source"]).lower()
                )

        _LOGGER.debug("Searching for fuel %s", fuel_type)
        locations = await self.find_fuel_locations_from_point(
            coordinates,
            radius,
            source_id)
        coros = [dynamic_build(l) for l in locations]
        await asyncio.gather(*coros)
        fuels: list = []
        for loc in locations:
            if loc["id"] not in self._accessed_sites:
                self._accessed_sites[loc["id"]] = loc["props"][PROP_FUEL_LOCATION_SOURCE]
            for fuel in loc["available_fuels"]:
                if (fuel == fuel_type) and (
                    loc["available_fuels"][fuel] > 0
                ):
                    fuels.append({
                        **loc,
                        "cost": loc["available_fuels"][fuel],
                        "distance": loc["distance"]
                    })
        return sorted(fuels, key=lambda item: item["cost"])

    @classmethod
    def create(cls,
               enabled_sources: list[str] = None,
               update_interval: timedelta = timedelta(days=1),
               country_code: str = "",
               configured_areas: list[dict] = None,
               timeout: timedelta = timedelta(seconds=30)
            ) -> 'FuelPrices':
        """Start an instance of fuel prices."""
        self = cls()
        self.configured_areas = configured_areas
        self.client_session = aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(
                use_dns_cache=True,
                ttl_dns_cache=360
            ),
            timeout=aiohttp.ClientTimeout(
                total=timeout.seconds
            )
        )
        if enabled_sources is not None:
            for src in enabled_sources:
                if str(src) not in SOURCE_MAP:
                    raise ValueError(f"Source {src} is not valid for this application.")
                self.configured_sources[src] = (
                    SOURCE_MAP.get(str(src))(update_interval=update_interval,
                                             client_session=self.client_session)
                )
        if enabled_sources is None:
            def_sources = {}
            if country_code != "":
                def_sources = COUNTRY_MAP.get(country_code.upper(), [])
            for src in def_sources:
                self.configured_sources[src] = (
                    SOURCE_MAP.get(str(src))(update_interval=update_interval,
                                             client_session=self.client_session)
                )

        return self



"""JET UK module."""

from pyfuelprices.sources.uk import CMAParser, FuelLocation

class JetUKSource(CMAParser):
    """JET uses the CMA parser."""

    _url = "https://jetlocal.co.uk/fuel_prices_data.json"
    provider_name = "jet"
    _timeout = 10
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None



"""Netherlands data const."""

API_BASE = "https://pg-app-hnf14cfy2xb2v9x9eueuchcd2xyetd.scalabl.cloud"

PECO_ONLINE_API = f"{API_BASE}/1/classes/farapret3"
PECO_APP_ID = "YueWcf0orjSz3IQmaT8yBNDTM5POP0mOU6EDyE3U"
PECO_APP_CLIENT_KEY = "ctPx9Ahrz9aaXhEvN0oWCzlX8FHX1cv3r7vZwxH8"

USER_AGENT = "Parse Android SDK API Level 34"
X_PARSE_OS_VERSION = "14"
X_PARSE_APP_DISPLAY_VERSION = "4.0"
X_PARSE_APP_BUILD_VERSION = "69"

PECO_STANDARD_HEADERS = {
    "X-Parse-Application-Id": PECO_APP_ID,
    "X-Parse-App-Build-Version": X_PARSE_APP_BUILD_VERSION,
    "X-Parse-App-Display-Version": X_PARSE_APP_DISPLAY_VERSION,
    "X-Parse-OS-Version": X_PARSE_OS_VERSION,
    "User-Agent": USER_AGENT,
    "X-Parse-Installation-Id": None,
    "X-Parse-Client-Key": PECO_APP_CLIENT_KEY,
    "Content-Type": "application/json"
}

PECO_QUERY_CLAUSE = '{\"Benzina_Regular\":{\"$gt\":0},\"DoarGPL\":{\"$ne\":1},\"Retea\":{\"$in\":[\"Gazprom\",\"Lukoil\",\"Mol\",\"OMV\",\"Petrom\",\"Rompetrol\",\"Socar\",\"BLKOil\",\"Ozana\",\"RST\"]}}'

PECO_FUEL_MAPPING = {
    "Benzina_Regular": "Gasoline Regular",
    "Benzina_Premium": "Gasoline Premium",
    "Motorina_Regular": "Diesel Regular",
    "Motorina_Premium": "Diesel Premium",
    "GPL": "GPL",
    "AdBlue": "AdBlue"
}



"""BP UK dataprovider."""

from pyfuelprices.const import DESKTOP_USER_AGENT
from pyfuelprices.fuel import Fuel
from pyfuelprices.sources.uk import CMAParser, FuelLocation

class BpUKSource(CMAParser):
    """BP UK uses the CMA parser."""

    _url = "https://www.bp.com/en_gb/united-kingdom/home/fuelprices/fuel_prices_data.json"
    provider_name = "bpuk"
    _headers = {
        "User-Agent": DESKTOP_USER_AGENT
    }
    _timeout = 60
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None



"""Austria spritpreisrechner data source."""

from datetime import datetime

import logging
import json

from pyfuelprices.const import (
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_SOURCE_ID,
    DESKTOP_USER_AGENT
)
from pyfuelprices.fuel_locations import FuelLocation, Fuel
from pyfuelprices.sources import Source

_LOGGER = logging.getLogger(__name__)

CONST_API_BASE = "https://www.spritpreisrechner.at/api"
CONST_API_SEARCH_REGIONS = (
    f"{CONST_API_BASE}/search/gas-stations/by-region?"
    "code={REG_CODE}&fuelType={FUEL_TYPE}&includeClosed=true&type={REG_TYPE}")
CONST_API_GET_REGIONS = (
    f"{CONST_API_BASE}/regions"
)
CONST_FUELS = ["DIE", "SUP", "GAS"]

class SpripreisrechnerATSource(Source):
    """Data source for spritpreisrechner."""
    provider_name = "spritpreisrechner"
    _headers = {
        "User-Agent": DESKTOP_USER_AGENT
    }
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None
    _regions = []

    async def _send_request(self, url):
        """Send a HTTP request to the API."""
        async with self._client_session.get(
            url,
            headers=self._headers
        ) as response:
            if response.ok:
                return await response.text()
            _LOGGER.error("Error sending request to %s: %s",
                            url,
                            response)

    async def _load_regions(self):
        """Load regions into memory."""
        self._regions = json.loads(
            await self._send_request(CONST_API_GET_REGIONS)
        )

    def _parse_raw_fuel_station(self, station, site_id) -> FuelLocation:
        """Parse a raw fuel station."""
        _LOGGER.debug("Parsing %s location ID %s", self.provider_name, site_id)
        loc = FuelLocation.create(
            site_id=site_id,
            name=station.get("name", f"Unknown {station['id']}"),
            address=station["location"]["address"],
            lat=station["location"]["latitude"],
            long=station["location"]["longitude"],
            brand="unavailable",
            available_fuels=[Fuel(
                fuel_type=station["prices"][0]["label"],
                cost=station["prices"][0]["amount"],
                props=station["prices"][0]
            )],
            postal_code=station["location"]["postalCode"],
            currency="EUR",
            props={
                PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                PROP_FUEL_LOCATION_SOURCE_ID: station["id"],
                PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True,
                "data": station
            }
        )
        loc.next_update = self.next_update + self.update_interval
        return loc

    def _update_fuel_station_prices(self, station, site_id):
        """Internal method to update station prices."""
        fuel = station["prices"][0]
        try:
            self.location_cache[site_id].get_fuel(fuel["label"]).update(
                fuel_type=fuel["label"],
                cost=fuel["amount"],
                props=fuel
            )
        except ValueError:
            self.location_cache[site_id].available_fuels.append(
                Fuel(
                    fuel_type=fuel["label"],
                    cost=fuel["amount"],
                    props=fuel
                )
            )
        self.location_cache[site_id].last_updated = datetime.now()
        self.location_cache[site_id].next_update = (
            self.next_update + self.update_interval
        )

    async def update(self, areas=None, force=None) -> list[FuelLocation]:
        """Custom update handler to query each region."""
        if datetime.now() > self.next_update:
            if len(self._regions) == 0:
                await self._load_regions()
            for region in self._regions:
                for fuel in CONST_FUELS:
                    url = CONST_API_SEARCH_REGIONS.format(
                        REG_CODE=str(region["code"]),
                        FUEL_TYPE=fuel,
                        REG_TYPE=str(region["type"])
                    )
                    response_raw = json.loads(await self._send_request(url))
                    for station in response_raw:
                        site_id = f"{self.provider_name}_{station['id']}"
                        if site_id in self.location_cache:
                            self._update_fuel_station_prices(station, site_id)
                            continue
                        loc = self._parse_raw_fuel_station(station, site_id)
                        self.location_cache[loc.id] = loc
                        continue
            return list(self.location_cache.values())

    async def parse_response(self, response) -> list[FuelLocation]:
        """Method not used."""

    def parse_fuels(self, fuels) -> list[Fuel]:
        """Method not used."""



"""Motor Fuel Group UK dataprovider."""

from pyfuelprices.sources.uk import CMAParser, FuelLocation

class MotorFuelGroupUKSource(CMAParser):
    """Motor Fuel Group UK uses the CMA parser."""

    _url = "https://fuel.motorfuelgroup.com/fuel_prices_data.json"
    provider_name = "motorfuelgroup"
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None



"""Asda UK module."""

from pyfuelprices.sources.uk import CMAParser, FuelLocation

class AsdaUKSource(CMAParser):
    """Asda uses the CMA parser."""

    _url = "https://storelocator.asda.com/fuel_prices_data.json"
    provider_name = "asda"
    _timeout = 10
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None



"""Rontec UK dataprovider."""

from pyfuelprices.sources.uk import CMAParser, FuelLocation

class RontecUKSource(CMAParser):
    """Rontec UK uses the same parses as CMA."""

    _url = "https://www.rontec-servicestations.co.uk/fuel-prices/data/fuel_prices_data.json"
    provider_name = "rontec"
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None



import re

import setuptools

with open('README.md', 'r', encoding='utf8') as readme_file:
    long_description = readme_file.read()

# Inspiration: https://stackoverflow.com/a/7071358/6064135
with open('pyfuelprices/_version.py', 'r', encoding='utf8') as version_file:
    version_groups = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file.read(), re.M)
    if version_groups:
        version = version_groups.group(1)
    else:
        raise RuntimeError('Unable to find version string!')

REQUIREMENTS = [
    # 'numpy == 1.26.0',
    # 'scikit-learn == 1.3.0',
    'these-united-states == 1.1.0.21',
    'geopy == 2.4.0',
]

DEV_REQUIREMENTS = [
    'bandit == 1.7.*',
    'black == 23.*',
    'build == 0.10.*',
    'flake8 == 6.*',
    'isort == 5.*',
    'mypy == 1.5.*',
    'pytest == 7.*',
    'pytest-cov == 4.*',
    'twine == 4.*',
]

setuptools.setup(
    name='pyfuelprices',
    version=version,
    description='A generic library to collect fuel prices of fuel stations around the world!',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/pantherale0/pyfuelprices',
    author='pantherale0',
    license='MIT',
    packages=setuptools.find_packages(),
    package_data={
        'pyfuelprices': [
            'py.typed',
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=REQUIREMENTS,
    extras_require={
        'dev': DEV_REQUIREMENTS,
    },
    entry_points={
        'console_scripts': [
            'pyfuelprices=pyfuelprices:main',
        ]
    },
    python_requires='>=3.8, <4',
)



"""FuelWatch data source for Australia."""

from datetime import datetime

import logging
import json

from pyfuelprices.const import (
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_SOURCE_ID,
    DESKTOP_USER_AGENT
)
from pyfuelprices.fuel_locations import FuelLocation, Fuel
from pyfuelprices.sources import Source

from .const import FUELWATCH_API_PRODUCTS, FUELWATCH_API_SITES

_LOGGER = logging.getLogger(__name__)

class FuelWatchSource(Source):
    """FuelWatch data source."""
    provider_name = "fuelwatch"
    _fuel_products: list[str] = []
    _headers = {
        "User-Agent": DESKTOP_USER_AGENT
    }
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None

    async def _send_request(self, url):
        """Send a HTTP request to the API and return the text."""
        async with self._client_session.get(
            url,
            headers=self._headers) as response:
            if response.ok:
                return await response.text()
            _LOGGER.error("Error sending request to %s: %s",
                            url,
                            response)

    async def get_fuel_products(self):
        """Retrieves available fuel products."""
        response_raw = await self._send_request(FUELWATCH_API_PRODUCTS)
        for product in json.loads(response_raw):
            if product["shortName"] not in self._fuel_products:
                self._fuel_products.append(product["shortName"])
        return self._fuel_products

    def _parse_raw_fuel_station(self, station, site_id) -> FuelLocation:
        """Convert an instance of a single fuel station into a FuelLocation."""
        _LOGGER.debug("Parsing FuelWatch location ID %s", site_id)
        loc = FuelLocation.create(
            site_id=site_id,
            name=station["siteName"],
            address=f"{station['address']['line1']} {station['address']['postCode']}",
            lat=station['address']['latitude'],
            long=station['address']['longitude'],
            brand=station['brandName'],
            available_fuels=[Fuel(
                fuel_type=station["productFuelType"],
                cost=station["product"].get("priceToday", 0),
                props=station["product"]
            )],
            postal_code=station['address']['postCode'],
            currency="",
            props={
                PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                PROP_FUEL_LOCATION_SOURCE_ID: station["id"],
                PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True,
                "data": station
            }
        )
        loc.next_update = self.next_update + self.update_interval
        return loc

    def _update_fuel_station_prices(self, station, site_id):
        """Internal method to update station prices."""
        try:
            self.location_cache[site_id].get_fuel(station["productFuelType"]).update(
                fuel_type=station["productFuelType"],
                cost=station["product"].get("priceToday", 0),
                props=station["product"]
            )
        except ValueError:
            self.location_cache[site_id].available_fuels.append(
                Fuel(
                    fuel_type=station["productFuelType"],
                    cost=station["product"].get("priceToday", 0),
                    props=station["product"]
                )
            )

    async def update(self, areas=None, force=None) -> list[FuelLocation]:
        """Custom update handler to look all products."""
        # _df_columns = ["lat", "long"]
        # _df_data = []
        if datetime.now() > self.next_update:
            if len(self._fuel_products) == 0:
                await self.get_fuel_products()
            for product in self._fuel_products:
                url = FUELWATCH_API_SITES.format(
                    FUELTYPE=product
                )
                response_raw = json.loads(await self._send_request(url))
                for station in response_raw:
                    site_id = f"{self.provider_name}_{station['id']}"
                    if site_id in self.location_cache:
                        self._update_fuel_station_prices(station, site_id)
                        continue
                    loc = self._parse_raw_fuel_station(station, site_id)
                    self.location_cache[loc.id] = loc
                    continue
            # for x in self.location_cache.values():
            #     _df_data.append([x.lat, x.long])
            # _df = pd.DataFrame(_df_data, columns=_df_columns)
            # self.location_tree = KDTree(_df[["lat", "long"]].values,
            #                             metric="euclidean")
            self.next_update += self.update_interval
            return list(self.location_cache.values())

    async def parse_response(self, response) -> list[FuelLocation]:
        """Method not used."""

    def parse_fuels(self, fuels) -> list[Fuel]:
        """Method not used."""



"""FuelSnoop data source for Australia (QLD)."""

from datetime import datetime
import logging
import json

from geopy import distance

from pyfuelprices.const import (
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_SOURCE_ID,
    PROP_AREA_LAT,
    PROP_AREA_LONG,
    PROP_AREA_RADIUS
)
from pyfuelprices.fuel_locations import FuelLocation, Fuel
from pyfuelprices.helpers import get_bounding_box
from pyfuelprices.sources import Source

from .const import FUELSNOOP_API_HEADERS, FUELSNOOP_API_SITES

_LOGGER = logging.getLogger(__name__)

class FuelSnoopSource(Source):
    """FuelSnoop data source."""
    provider_name = "fuelsnoop"
    _fuel_products: list[str] = []
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None

    async def _send_request(self, url, body):
        """Send a HTTP request to the API and return the text."""
        async with self._client_session.post(
            url,
            headers=FUELSNOOP_API_HEADERS,
            json=body) as response:
            if response.ok:
                return await response.text()
            _LOGGER.error("Error sending request to %s: %s",
                            url,
                            response)

    async def search_sites(self, coordinates, radius: float) -> list[dict]:
        """Return all available sites within a given radius."""
        # first query the API to populate cache / update data in case this data is unavailable.
        await self.update(
            areas=[{
                PROP_AREA_LAT: coordinates[0],
                PROP_AREA_LONG: coordinates[1],
                PROP_AREA_RADIUS: radius
            }],
            force=True
        )
        locations = []
        for site in self.location_cache.values():
            dist = distance.distance(coordinates,
                                 (
                                    site.lat,
                                    site.long
                                )).miles
            if dist < radius:
                locations.append({
                    **site.__dict__(),
                    "distance": dist
                })
        return locations

    async def update(self, areas=None, force=None) -> list[FuelLocation]:
        """Custom update handler to look all products."""
        self._configured_areas=[] if areas is None else areas
        if datetime.now() > self.next_update or force:
            for area in self._configured_areas:
                _LOGGER.debug("Searching FuelSnoop for FuelLocations at area %s",
                              area)
                bbox = get_bounding_box(
                    area[PROP_AREA_LAT],
                    area[PROP_AREA_LONG],
                    area[PROP_AREA_RADIUS]
                )
                response_raw = await self._send_request(
                    url=FUELSNOOP_API_SITES,
                    body={
                        "min_lng": bbox.lon_min,
                        "min_lat": bbox.lat_min,
                        "max_lng": bbox.lon_max,
                        "max_lat": bbox.lat_max,
                        "brand_names": []
                    }
                )
                if response_raw is not None:
                    await self.parse_response(json.loads(response_raw))
            self.next_update += self.update_interval
            return list(self.location_cache.values())

    async def parse_response(self, response) -> list[FuelLocation]:
        for station in response:
            await self.parse_raw_fuel_station(station=station)
        return list(self.location_cache.values())

    async def parse_raw_fuel_station(self, station) -> FuelLocation:
        """Convert an instance of a single fuel station into a FuelLocation."""
        site_id = f"{self.provider_name}_{station['id']}"
        _LOGGER.debug("Parsing FuelSnoop location ID %s", site_id)
        loc = FuelLocation.create(
            site_id=site_id,
            name=station["site_name"],
            address=station['address'],
            lat=station['lat'],
            long=station['lng'],
            brand=station['brand_name'],
            available_fuels=self.parse_fuels(station["prices"]),
            postal_code=station['postcode'],
            currency="AUD",
            props={
                PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                PROP_FUEL_LOCATION_SOURCE_ID: station["id"],
                PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True,
                "data": station
            }
        )
        loc.next_update = self.next_update + self.update_interval
        if site_id not in self.location_cache:
            self.location_cache[site_id] = loc
        else:
            await self.location_cache[site_id].update(loc)
        return self.location_cache[site_id]

    def parse_fuels(self, fuels: dict) -> list[Fuel]:
        output = []
        for k in fuels:
            f = fuels[k]
            if f["price"] > 999:
                continue # ignore this fuel as not available at this stn
            output.append(Fuel(
                fuel_type=k,
                cost=f["price"],
                props=f
            ))
        return output



"""Python Fuel Prices consts."""

PROP_FUEL_LOCATION_DYNAMIC_BUILD = "build_on_access"
PROP_FUEL_LOCATION_SOURCE = "source"
PROP_FUEL_LOCATION_SOURCE_ID = "source_id"
PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP = "prevent_cache_cleanup"

PROP_AREA_LAT = "lat"
PROP_AREA_LONG = "long"
PROP_AREA_RADIUS = "radius"


ANDROID_USER_AGENT = "Dalvik/2.1.0 (Linux; U; Android 13; Pixel 4 XL Build/TQ3A.230705.001.B4)"
DESKTOP_USER_AGENT = "Mozilla/5.0 (X11; Linux x86_64; rv:120.0) Gecko/20100101 Firefox/120.0"



"""AUS Fuel Data Consts."""

FUELWATCH_API_BASE = "https://www.fuelwatch.wa.gov.au/api/"
FUELWATCH_API_PRODUCTS = f"{FUELWATCH_API_BASE}products"
FUELWATCH_API_SITES = (
    f"{FUELWATCH_API_BASE}sites"
    "?fuelType={FUELTYPE}"
)

FUELSNOOP_API_BASE = "https://jqdyvthpvgnvlojefpav.supabase.co/rest/v1/rpc/"
FUELSNOOP_API_API_TOKEN = (
    "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
    "eyJpc3MiOiJzdXBhYmFzZSIsInJlZiI6ImpxZH"
    "l2dGhwdmdudmxvamVmcGF2Iiwicm9sZSI6ImFu"
    "b24iLCJpYXQiOjE3MDEwODM2MzksImV4cCI6Mj"
    "AxNjY1OTYzOX0.7fEHEq5g3OFLBSyzuOObdJLZ"
    "NlqFyVJPoYre2fYzN0E"
)
FUELSNOOP_API_HEADERS = {
    "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:127.0) Gecko/20100101 Firefox/127.0",
    "Accept": "application/json",
    "Accept-Language": "en-GB,en;q=0.5",
    "Accept-Encoding": "gzip, deflate, br, zstd",
    "Referer": "https://www.fuelsnoop.com.au/",
    "apikey": FUELSNOOP_API_API_TOKEN,
    "authorization": f"Bearer {FUELSNOOP_API_API_TOKEN}",
    "content-profile": "public",
    "x-client-info": "supabase-ssr/0.0.10",
    "Origin": "https://www.fuelsnoop.com.au",
    "Connection": "keep-alive",
    "Sec-Fetch-Dest": "empty",
    "Sec-Fetch-Mode": "cors",
    "Sec-Fetch-Site": "cross-site",
    "Priority": "u=4",
    "Pragma": "no-cache",
    "Cache-Control": "no-cache",
    "TE": "trailers"
}
FUELSNOOP_API_SITES = f"{FUELSNOOP_API_BASE}sites_in_view"
